@@ -1,216 +1,19 @@
 package Net::SSH2;
 
-our $VERSION = '0.59_04';
+our $VERSION = '0.59_05';
 
 use 5.006;
 use strict;
 use warnings;
 use Carp;
 
-require Exporter;
-use AutoLoader;
+require Net::SSH2::Constants;
 
 use Socket;
 use IO::File;
 use File::Basename;
 use Errno;
 
-use base 'Exporter';
-
-# constants
-
-my @EX_callback = qw(
-        LIBSSH2_CALLBACK_DEBUG
-        LIBSSH2_CALLBACK_DISCONNECT
-        LIBSSH2_CALLBACK_IGNORE
-        LIBSSH2_CALLBACK_MACERROR
-        LIBSSH2_CALLBACK_X11
-);
-
-my @EX_channel = qw(
-        LIBSSH2_CHANNEL_EXTENDED_DATA_IGNORE
-        LIBSSH2_CHANNEL_EXTENDED_DATA_MERGE
-        LIBSSH2_CHANNEL_EXTENDED_DATA_NORMAL
-);
-
-my @EX_socket = qw(
-        LIBSSH2_SOCKET_BLOCK_INBOUND
-        LIBSSH2_SOCKET_BLOCK_OUTBOUND
-);
-
-my @EX_trace = qw(
-        LIBSSH2_TRACE_TRANS
-        LIBSSH2_TRACE_KEX
-        LIBSSH2_TRACE_AUTH
-        LIBSSH2_TRACE_CONN
-        LIBSSH2_TRACE_SCP
-        LIBSSH2_TRACE_SFTP
-        LIBSSH2_TRACE_ERROR
-        LIBSSH2_TRACE_PUBLICKEY
-        LIBSSH2_TRACE_SOCKET
-);
-
-my @EX_error = qw(
-        LIBSSH2_ERROR_ALLOC
-        LIBSSH2_ERROR_BANNER_NONE
-        LIBSSH2_ERROR_BANNER_SEND
-        LIBSSH2_ERROR_CHANNEL_CLOSED
-        LIBSSH2_ERROR_CHANNEL_EOF_SENT
-        LIBSSH2_ERROR_CHANNEL_FAILURE
-        LIBSSH2_ERROR_CHANNEL_OUTOFORDER
-        LIBSSH2_ERROR_CHANNEL_PACKET_EXCEEDED
-        LIBSSH2_ERROR_CHANNEL_REQUEST_DENIED
-        LIBSSH2_ERROR_CHANNEL_UNKNOWN
-        LIBSSH2_ERROR_CHANNEL_WINDOW_EXCEEDED
-        LIBSSH2_ERROR_DECRYPT
-        LIBSSH2_ERROR_FILE
-        LIBSSH2_ERROR_HOSTKEY_INIT
-        LIBSSH2_ERROR_HOSTKEY_SIGN
-        LIBSSH2_ERROR_INVAL
-        LIBSSH2_ERROR_INVALID_MAC
-        LIBSSH2_ERROR_INVALID_POLL_TYPE
-        LIBSSH2_ERROR_KEX_FAILURE
-        LIBSSH2_ERROR_KEY_EXCHANGE_FAILURE
-        LIBSSH2_ERROR_METHOD_NONE
-        LIBSSH2_ERROR_METHOD_NOT_SUPPORTED
-        LIBSSH2_ERROR_PASSWORD_EXPIRED
-        LIBSSH2_ERROR_PROTO
-        LIBSSH2_ERROR_PUBLICKEY_UNRECOGNIZED
-        LIBSSH2_ERROR_PUBLICKEY_UNVERIFIED
-        LIBSSH2_ERROR_REQUEST_DENIED
-        LIBSSH2_ERROR_SCP_PROTOCOL
-        LIBSSH2_ERROR_SFTP_PROTOCOL
-        LIBSSH2_ERROR_SOCKET_DISCONNECT
-        LIBSSH2_ERROR_SOCKET_NONE
-        LIBSSH2_ERROR_SOCKET_SEND
-        LIBSSH2_ERROR_SOCKET_TIMEOUT
-        LIBSSH2_ERROR_TIMEOUT
-        LIBSSH2_ERROR_ZLIB
-        LIBSSH2_ERROR_EAGAIN
-);
-
-my @EX_hash = qw(
-        LIBSSH2_HOSTKEY_HASH_MD5
-        LIBSSH2_HOSTKEY_HASH_SHA1
-);
-
-my @EX_method = qw(
-        LIBSSH2_METHOD_COMP_CS
-        LIBSSH2_METHOD_COMP_SC
-        LIBSSH2_METHOD_CRYPT_CS
-        LIBSSH2_METHOD_CRYPT_SC
-        LIBSSH2_METHOD_HOSTKEY
-        LIBSSH2_METHOD_KEX
-        LIBSSH2_METHOD_LANG_CS
-        LIBSSH2_METHOD_LANG_SC
-        LIBSSH2_METHOD_MAC_CS
-        LIBSSH2_METHOD_MAC_SC
-);
-
-my @EX_fxf = qw(
-        LIBSSH2_FXF_APPEND
-        LIBSSH2_FXF_CREAT
-        LIBSSH2_FXF_EXCL
-        LIBSSH2_FXF_READ
-        LIBSSH2_FXF_TRUNC
-        LIBSSH2_FXF_WRITE
-);
-
-my @EX_fx = qw(
-        LIBSSH2_FX_BAD_MESSAGE
-        LIBSSH2_FX_CONNECTION_LOST
-        LIBSSH2_FX_DIR_NOT_EMPTY
-        LIBSSH2_FX_EOF
-        LIBSSH2_FX_FAILURE
-        LIBSSH2_FX_FILE_ALREADY_EXISTS
-        LIBSSH2_FX_INVALID_FILENAME
-        LIBSSH2_FX_INVALID_HANDLE
-        LIBSSH2_FX_LINK_LOOP
-        LIBSSH2_FX_LOCK_CONFlICT
-        LIBSSH2_FX_NOT_A_DIRECTORY
-        LIBSSH2_FX_NO_CONNECTION
-        LIBSSH2_FX_NO_MEDIA
-        LIBSSH2_FX_NO_SPACE_ON_FILESYSTEM
-        LIBSSH2_FX_NO_SUCH_FILE
-        LIBSSH2_FX_NO_SUCH_PATH
-        LIBSSH2_FX_OK
-        LIBSSH2_FX_OP_UNSUPPORTED
-        LIBSSH2_FX_PERMISSION_DENIED
-        LIBSSH2_FX_QUOTA_EXCEEDED
-        LIBSSH2_FX_UNKNOWN_PRINCIPLE
-        LIBSSH2_FX_WRITE_PROTECT
-);
-
-my @EX_sftp = qw(
-        LIBSSH2_SFTP_ATTR_ACMODTIME
-        LIBSSH2_SFTP_ATTR_EXTENDED
-        LIBSSH2_SFTP_ATTR_PERMISSIONS
-        LIBSSH2_SFTP_ATTR_SIZE
-        LIBSSH2_SFTP_ATTR_UIDGID
-        LIBSSH2_SFTP_LSTAT
-        LIBSSH2_SFTP_OPENDIR
-        LIBSSH2_SFTP_OPENFILE
-        LIBSSH2_SFTP_PACKET_MAXLEN
-        LIBSSH2_SFTP_READLINK
-        LIBSSH2_SFTP_REALPATH
-        LIBSSH2_SFTP_RENAME_ATOMIC
-        LIBSSH2_SFTP_RENAME_NATIVE
-        LIBSSH2_SFTP_RENAME_OVERWRITE
-        LIBSSH2_SFTP_SETSTAT
-        LIBSSH2_SFTP_STAT
-        LIBSSH2_SFTP_SYMLINK
-        LIBSSH2_SFTP_TYPE_BLOCK_DEVICE
-        LIBSSH2_SFTP_TYPE_CHAR_DEVICE
-        LIBSSH2_SFTP_TYPE_DIRECTORY
-        LIBSSH2_SFTP_TYPE_FIFO
-        LIBSSH2_SFTP_TYPE_REGULAR
-        LIBSSH2_SFTP_TYPE_SOCKET
-        LIBSSH2_SFTP_TYPE_SPECIAL
-        LIBSSH2_SFTP_TYPE_SYMLINK
-        LIBSSH2_SFTP_TYPE_UNKNOWN
-        LIBSSH2_SFTP_VERSION
-);
-
-my @EX_disconnect = qw(
-        SSH_DISCONNECT_AUTH_CANCELLED_BY_USER
-        SSH_DISCONNECT_BY_APPLICATION
-        SSH_DISCONNECT_COMPRESSION_ERROR
-        SSH_DISCONNECT_CONNECTION_LOST
-        SSH_DISCONNECT_HOST_KEY_NOT_VERIFIABLE
-        SSH_DISCONNECT_HOST_NOT_ALLOWED_TO_CONNECT
-        SSH_DISCONNECT_ILLEGAL_USER_NAME
-        SSH_DISCONNECT_KEY_EXCHANGE_FAILED
-        SSH_DISCONNECT_MAC_ERROR
-        SSH_DISCONNECT_NO_MORE_AUTH_METHODS_AVAILABLE
-        SSH_DISCONNECT_PROTOCOL_ERROR
-        SSH_DISCONNECT_PROTOCOL_VERSION_NOT_SUPPORTED
-        SSH_DISCONNECT_RESERVED
-        SSH_DISCONNECT_SERVICE_NOT_AVAILABLE
-        SSH_DISCONNECT_TOO_MANY_CONNECTIONS
-);
-
-our %EXPORT_TAGS = (
-    all        => [
-        @EX_callback, @EX_channel, @EX_error, @EX_socket, @EX_trace, @EX_hash,
-        @EX_method, @EX_fx, @EX_fxf, @EX_sftp, @EX_disconnect,
-    ],
-    # ssh
-    callback   => \@EX_callback,
-    channel    => \@EX_channel,
-    error      => \@EX_error,
-    socket     => \@EX_socket,
-    trace      => \@EX_trace,
-    hash       => \@EX_hash,
-    method     => \@EX_method,
-    disconnect => \@EX_disconnect,
-    # sftp
-    fx         => \@EX_fx,
-    fxf        => \@EX_fxf,
-    sftp       => \@EX_sftp,
-);
-
-our @EXPORT_OK = @{$EXPORT_TAGS{all}};
-
 # load IO::Socket::IP when available, otherwise fallback to IO::Socket::INET.
 
 my $socket_class = do {
@@ -290,12 +93,12 @@ sub connect {
         }
     }
 
-    my ($remote_hostname, $remote_port);
+    my ($hostname, $port);
     if (@_ == 2) {
-        $remote_hostname = $_[0];
-        $remote_port = getservbyname($_[1] || 'ssh', 'tcp') || 22;
-        $sock = $socket_class->new( PeerHost => $remote_hostname,
-                                    PeerPort => $remote_port,
+        $hostname = $_[0];
+        $port = getservbyname($_[1] || 'ssh', 'tcp') || 22;
+        $sock = $socket_class->new( PeerHost => $hostname,
+                                    PeerPort => $port,
                                     Blocking => $self->blocking,
                                     Timeout => $self->timeout );
         unless ($sock) {
@@ -321,14 +124,14 @@ sub connect {
 
     {
         local ($@, $SIG{__DIE__});
-        $remote_port = eval { $sock->peerport }
-            unless defined $remote_port;
-        $remote_hostname = eval { $sock->peername } || 22
-            unless defined $remote_hostname;
+        $port = eval { $sock->peerport }
+            unless defined $port;
+        $hostname = eval { $sock->peername } || 22
+            unless defined $hostname;
     }
 
     # pass it in, do protocol
-    return $self->_startup($fd, $sock, $remote_hostname, $remote_port);
+    return $self->_startup($fd, $sock, $hostname, $port);
 
  error:
     unless (defined wantarray) {
@@ -483,15 +286,23 @@ sub auth_password_interact {
     return $rc;
 }
 
-sub check_remote_hostkey {
-    my ($self, $path, $policy) = @_;
+sub check_hostkey {
+    my ($self, $policy, $path, $comment) = @_;
+    my $cb;
+    if (not defined $policy) {
+        $policy = LIBSSH2_HOSTKEY_POLICY_STRICT();
+    }
+    elsif (ref $policy eq 'CODE') {
+        $cb = $policy;
+    }
+    else {
+        $policy =  _parse_constant(HOSTKEY_POLICY => $policy);
+    }
 
-    return 1 if $policy eq 'advisory'; # user doesn't care!
-
-    my $remote_hostname = $self->remote_hostname;
-    croak("remote_hostname unknown: in order to use check_remote_hostkey the peer host name ".
+    my $hostname = $self->hostname;
+    croak("hostname unknown: in order to use check_hostkey the peer host name ".
           "must be given (or discoverable) at connect time")
-        unless defined $remote_hostname;
+        unless defined $hostname;
 
     unless (defined $path) {
         my $home = $ENV{HOME} || (getpwuid($<))[7];
@@ -503,36 +314,59 @@ sub check_remote_hostkey {
         $path = File::Spec->catfile($home, '.ssh', 'known_hosts');
     }
 
-    my $kh = $self->known_hosts or return;
-    my $n_ent = $kh->readfile($path);
-    defined $n_ent or return;
+    my ($check, $key, $type, $flags);
+    my $kh = $self->known_hosts;
+    if ($kh and defined $kh->readfile($path)) {
 
-    my ($key, $type) = $self->remote_hostkey;
-    my $flags = ( LIBSSH2_KNOWNHOST_TYPE_PLAIN() |
-                  LIBSSH2_KNOWNHOST_KEYENC_RAW() |
-                  (($type + 1) << LIBSSH2_KNOWNHOST_KEY_SHIFT()) );
+        ($key, $type) = $self->remote_hostkey;
+        $flags = ( LIBSSH2_KNOWNHOST_TYPE_PLAIN() |
+                   LIBSSH2_KNOWNHOST_KEYENC_RAW() |
+                   (($type + 1) << LIBSSH2_KNOWNHOST_KEY_SHIFT()) );
 
-    my $check = $kh->check($remote_hostname, $self->remote_port, $key, $flags);
-    $check == LIBSSH2_KNOWNHOST_CHECK_MATCH() and return 1;
+        $check = $kh->check($hostname, $self->port, $key, $flags);
+        $check == LIBSSH2_KNOWNHOST_CHECK_MATCH() and return "00";
+    }
+    else {
+        $check = LIBSSH2_KNOWNHOST_CHECK_FAILURE();
+    }
+
+    if ($cb) {
+        my $ok = $cb->($self, $check, $comment);
+        $ok or $self->_set_error(LIBSSH2_ERROR_KNOWN_HOSTS(), 'Host key verification failed');
+        return $ok;
+    }
+
+    return $check
+        if $policy == LIBSSH2_HOSTKEY_POLICY_ADVISORY(); # user doesn't care!
 
     if ($check == LIBSSH2_KNOWNHOST_CHECK_NOTFOUND()) {
-        if ($policy eq 'ask') {
-            my $fp = unpack 'H*', $self->hostkey_hash(LIBSSH2_HOSTKEY_HASH_SHA1());
-            my $yes = $self->_ask_user("The authenticity of host '$remote_hostname' can't be established.\n" .
-                                       "key fingerprint is SHA1:$fp.\n" .
-                                       "Are you sure you want to continue connecting (yes/no)? ", 1);
-            if (lc $yes eq 'yes') {
-                return 1;
+        $self->_set_error(LIBSSH2_ERROR_KNOWN_HOSTS(), 'Unable to verify host key, host not found');
+        unless ($policy == LIBSSH2_HOSTKEY_POLICY_TOFU()) {
+            if ($policy == LIBSSH2_HOSTKEY_POLICY_ASK()) {
+                my $fp = unpack 'H*', $self->hostkey_hash(LIBSSH2_HOSTKEY_HASH_SHA1());
+                my $yes = $self->_ask_user("The authenticity of host '$hostname' can't be established.\n" .
+                                           "key fingerprint is SHA1:$fp.\n" .
+                                           "Are you sure you want to continue connecting (yes/no)? ", 1);
+                unless ($yes =~ /^y(es)?$/i) {
+                    $self->_set_error(LIBSSH2_ERROR_KNOWN_HOSTS(), 'Host key verification failed: user did not accept the key');
+                    return undef;
+                }
             }
         }
-        elsif ($policy eq 'tofu') {
-            return 1;
-        }
-    }
-    # else policy is 'strict' or the key doesn't match the one in known_hosts
 
-    $self->_set_error(LIBSSH2_ERROR_KNOWN_HOSTS(), 'Unable to verify remote host key');
-    ()
+        $comment = '(Net::SSH2)' unless defined $comment;
+        # we ignore errors here, that is the usual SSH client behaviour
+        $kh->add($hostname, $self->port, $key, $comment, $flags) and
+            $kh->writefile($path);
+
+        return $check;
+    }
+
+    $self->_set_error(LIBSSH2_ERROR_KNOWN_HOSTS(), 'Host key verification failed: '.
+                      ( ($check == LIBSSH2_KNOWNHOST_CHECK_NOTFOUND())
+                        ? "key not found in '$path'"
+                        : "unable to perform the check"));
+    return undef;
 }
 
 sub scp_get {
@@ -770,14 +604,16 @@ Net::SSH2 - Support for the SSH 2 protocol via libssh2.
 
   my $ssh2 = Net::SSH2->new();
 
-  $ssh2->connect('example.com') or die $ssh2->error;
+  $ssh2->connect('example.com') or $ssh2->die_with_error;
+
+  $ssh->check_hostkey('ask') or $ssh2->die_with_error;
 
   if ($ssh2->auth_keyboard('fizban')) {
       my $chan = $ssh2->channel();
       $chan->exec('program');
 
       my $sftp = $ssh2->sftp();
-      my $fh = $sftp->open('/etc/passwd') or die;
+      my $fh = $sftp->open('/etc/passwd') or $ssh->die_with_error;
       print $_ while <$fh>;
   }
 
@@ -787,11 +623,22 @@ C<Net::SSH2> is a perl interface to the libssh2 (L<http://www.libssh2.org>)
 library.  It supports the SSH2 protocol (there is no support for SSH1)
 with all of the key exchanges, ciphers, and compression of libssh2.
 
+Even if the module can be compiled and linked against very old
+versions of the library, nothing below 1.5.0 should really be used
+(older versions were quite buggy and unreliable) and version 1.7.0 or
+later is recommended.
+
 =head2 Error handling
 
 Unless otherwise indicated, methods return a true value on success and
 false on failure; use the error method to get extended error information.
 
+Methods in Net::SSH2 not backed by libssh2 functions
+(i.e. C<check_hostkey> or SCP related methods) require libssh2 1.7.0
+or later in order to set the error state or, in other words, when an
+older version of the library is used, after any of those methods fails
+C<error> would not return the real code but just some bogus result.
+
 =head2 Typical usage
 
 The typical usage order is as follows:
@@ -871,6 +718,10 @@ Key hash constants.
 
 Disconnect type constants.
 
+=item policy
+
+Policies for method C<check_hostkey>.
+
 =back
 
 SFTP constants:
@@ -1137,6 +988,15 @@ Send a clean disconnect message to the remote server.  Default values are empty
 strings for description and language, and C<SSH_DISCONNECT_BY_APPLICATION> for
 the reason.
 
+=head2 hostname
+
+The name of the remote host given at connect time or retrieved from
+the TCP connection.
+
+=head2 port
+
+The port number or the remote SSH server.
+
 =head2 hostkey_hash ( hash type )
 
 Returns a hash of the host key; note that the key is raw data and may contain
@@ -1159,14 +1019,50 @@ Returns the public key of the remote host and its type which is one of
 C<LIBSSH2_HOSTKEY_TYPE_RSA>, C<LIBSSH2_HOSTKEY_TYPE_DSS>, or
 C<LIBSSH2_HOSTKEY_TYPE_UNKNOWN>.
 
-=head2 check_remote_hostkey( [known_hosts_path, [policy]] )
+=head2 check_hostkey( [policy, [known_hosts_path [, comment] ] ] )
 
-Looks for the remote host key in the given file.
+Looks for the remote host key inside the given known host file
+(defaults to C<~/.ssh/known_hosts>).
 
-The path to the file containing the known host keys defaults to
-C<~/.ssh/known_hosts>.
+This method returns undef if the check fails or the result of the call
+to C<Net::SSH2::KnownHost::check>.
 
-Currently, the C<policy> argument is ignored.
+The accepted policies are as follows:
+
+=over 4
+
+=item LIBSSH2_HOSTKEY_POLICY_STRICT
+
+Only host keys already present in the known hosts file are accepted.
+
+=item LIBSSH2_HOSTKEY_POLICY_ASK
+
+If the host key is not present in the known hosts file, the user is
+asked if it should be accepted or not.
+
+If accepted, the key is added to the known host file with the given
+comment.
+
+=item LIBSSH2_HOSTKEY_POLICY_TOFU
+
+Trust On First Use: if the host key is not present in the known hosts
+file, it is added there and accepted.
+
+=item LIBSSH2_HOSTKEY_POLICY_ADVISORY
+
+The key is always accepted, but it is never saved into the known host
+file.
+
+=item $callback
+
+If a reference to a subroutine is given, it is called when the key is
+not present in the known hosts file or a different key is found. The
+arguments passed to the callback are the session object, the matching
+error (C<LIBSSH2_KNOWNHOST_CHECK_FAILURE>,
+C<LIBSSH2_KNOWNHOST_CHECK_NOTFOUND> or
+C<LIBSSH2_KNOWNHOST_CHECK_MISMATCH>) and the comment.
+
+=back
 
 =head2 auth_list ( [username] )
 
